@@ -61,11 +61,7 @@ impl<'source> ParsingContext<'source> {
                 let ty = parser.module.types.insert(
                     Type {
                         name: Some(ty_name.clone()),
-                        inner: TypeInner::Struct {
-                            top_level: false,
-                            members,
-                            span,
-                        },
+                        inner: TypeInner::Struct { members, span },
                     },
                     meta,
                 );
